@@ -316,6 +316,91 @@ fixedeffects.pvargmm <- function(model, Only_Non_NA_rows = TRUE, ...){
 
 }
 
+#' Extracting Level Residuals
+#' @param model Model
+#' @param ... Further arguments passed to or from other methods
+#' @export
+#' @examples
+#' data("ex1_dahlberg_data")
+#' residuals_level(ex1_dahlberg_data)
+
+residuals_level <- function(model, ...) UseMethod("residuals_level")
+
+#' @rdname residuals_level
+#' @export
+
+residuals_level.pvargmm <- function(model, ...){
+  
+  #model <-  test_Q_for_endo
+  lags <- c(1:model$lags)
+  
+  Y_mean <- c(model$dependent_vars)
+  
+  # Improve next line:
+  X_mean_1 <- unique(matrix(mapply(function(i) mapply(function(j)
+    paste("lag", lags[j], "_", model$dependent_vars, sep = ""),  1:length(lags)), 1:length(model$dependent_vars)), ncol = 1))
+  
+  X_mean_2 <- c(if(!is.null(model$predet_vars)){model$predet_vars}, if(!is.null(model$exog_vars)){model$exog_vars})
+  
+  X_mean <- c(X_mean_1, if(!is.null(X_mean_2)){X_mean_2}, if(model$system_instruments==TRUE){"const"})
+  
+  # Add a column with 1 if system instruments are TRUE.
+  if (model$system_instruments==TRUE){
+    model$Set_Vars$const <- 1
+  }
+  
+  Set_Vars_reduced <- subset(model$Set_Vars, select = c("category", "period" , Y_mean, X_mean))
+  
+  
+  if (model$steps == "onestep"){Beta <- model$first_step}
+  if (model$steps == "twostep"){Beta <- model$second_step}
+  
+  # Get fixed effect:
+  fixed_eff_list <- fixedeffects(model)
+  
+  # Calculate vector of residuals for each panel category separately:
+  
+  list.residuals <- list()
+  
+  vector.categories <- unique(Set_Vars_reduced$category)
+  
+  for (i0 in 1:length(vector.categories)){
+    
+    # We accept NAs in the first p-lags:
+    # Get data for i
+    Set_Vars_reduced.i <- subset(Set_Vars_reduced, 
+                                 Set_Vars_reduced$category == vector.categories[i0])
+    
+    # Get fixed effects for i
+    fixed_eff_list.i <- fixed_eff_list[[i0]]
+    
+    # Dimension: [nof_dependent] x [T]
+    big.Y <- t(Set_Vars_reduced.i[,Y_mean])
+    
+    # Dimension: [nof_dependent] x [nof RHS variables plus 1 row for the fixed effect]
+    Big.Beta <- cbind(fixed_eff_list.i, Beta)
+    
+    # Dimension: [nof RHS variables plus 1 row for the fixed effect] x [T]
+    Big.X <- rbind(1, t(Set_Vars_reduced.i[,X_mean]))
+    
+    # Dimension: [nof RHS variables] x [T]
+    list.residuals[[i0]] <- big.Y - Big.Beta %*% Big.X
+    colnames(list.residuals[[i0]]) <- Set_Vars_reduced.i$period
+  }
+  
+  categories <- unique(Set_Vars_reduced$category)
+  # Name the list entries with the category identifier:
+  names(list.residuals) <- categories
+  
+  # Make a table with first two columns as category, period similar to Set_Vars
+  df_residuals <- mapply(function(i) t(list.residuals[[i]]), 1:length(list.residuals), SIMPLIFY = FALSE)
+  
+  df_residuals <- do.call("rbind", mapply(function(i) data.frame(category = categories[i], 
+                                                                 period = row.names(df_residuals[[i]]), df_residuals[[i]]), 1:length(df_residuals), SIMPLIFY = FALSE))
+  row.names(df_residuals) <- NULL
+  
+  return(df_residuals)  
+}  
 
 #' @export
 vcov.pvargmm <- function(object, ...){
